 /*
  * Unitex
  *
  * Copyright (C) 2001-2004 Universit de Marne-la-Valle <unitex@univ-mlv.fr>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  * 
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
  *
  */

#include "autalmot.h"
#include "symbol_op.h"
#include "stack.h"
#include "fst_file.h"
#include "utils.h"

static void add_limphrase(autalmot_t * A) {

  static unichar S[] = { '{', 'S', '}', 0 };

  int idx = language_add_form(LANG, S);

  symbol_t * LIM = symbol_PUNC_new(LANG, idx);

  int initBis   = autalmot_add_state(A);
  int nouvFinal = autalmot_add_state(A, AUT_TERMINAL);

  A->states[initBis].trans  = A->states[A->initials[0]].trans;
  A->states[A->initials[0]].trans = NULL;

  A->states[initBis].flags = A->states[A->initials[0]].flags & ~(AUT_INITIAL);
  A->states[A->initials[0]].flags = AUT_INITIAL;

  autalmot_add_trans(A, A->initials[0], LIM, initBis);


  for (int q = 1; q < A->nbstates - 2; q++) {
    if (autalmot_is_final(A, q)) {
      autalmot_add_trans(A, q, LIM, nouvFinal);
      autalmot_unset_terminal(A, q);
    }
  }

  symbol_delete(LIM);
}

void print_match(stack_type * stack, FILE * f) {

  for (int i = 0; i < stack->head; i++) {
    symbol_dump((symbol_t *) stack->data[i], f); fprintf(f, " ");
  }
  fprintf(f, "\n");
}


static void locate(autalmot_t * sentence, int q1, autalmot_t * pattern, int q2,
                   int * nbm, stack_type * stack, FILE * f) {

  transition_t * t1, * t2;

  if (autalmot_is_final(pattern, q2)) { (*nbm)++; print_match(stack, f); }
  
  for (t1 = sentence->states[q1].trans; t1; t1 = t1->next) {
  
    for (t2 = pattern->states[q2].trans; t2; t2 = t2->next) {
    
      if (symbol_in_symbol(t1->label, t2->label)) {
        stack_push(stack, t1->label);
        locate(sentence, t1->to, pattern, t2->to, nbm, stack, f);
        stack_pop(stack);
      }
    }
  }
}



int autalmot_locate_pattern(autalmot_t * sentence, autalmot_t * pattern, FILE * f) {

  int nbmatches = 0;
  stack_type * stack = stack_new();

  for (int i = 0; i < sentence->nbstates; i++) {
    locate(sentence, i, pattern, 0, & nbmatches, stack, f); 
  }
  
  printf("%d matches\n", nbmatches);

  stack_delete(stack);
  
  return nbmatches;
}



void usage() {
  printf("usage: bla txt -p pattern -l LANG\n");
}

int main(int argc, char ** argv) {
  

  char * txtname = NULL, * langname = NULL, * patternname = NULL;

  argv++, argc--;

  if (argc == 0) { usage(); return 0; }

  while (argc) {

    if (**argv != '-') {

      txtname = *argv;

    } else {

      if (strcmp(*argv, "-h") == 0) {

	usage();
        return 0;

      } else if (strcmp(*argv, "-p") == 0) {

	*argv++, argc--;
	if (argc == 0) { die("'-p' needs an additionnal argument\n"); }

	patternname = *argv;
      
      } else if (strcmp(*argv, "-l") == 0) { 

	argv++, argc--;
	if (argc == 0) { die("-l needs an arg\n"); }
	langname = *argv;
      
      } else { die("unknow argument '%s'\n", *argv); }  
    }

    *argv++, argc--;
  }

  if (txtname == NULL) { die("no text automaton specified\n"); }
  if (langname == NULL) { die("no text language specified\n"); }
  if (patternname == NULL) { die("no pattern specified\n"); }


  printf("loading %s langage definition ...\n", langname);

  language_t * lang = language_load(langname);
  set_current_language(lang);


  fst_file_in_t * txtin = fst_file_in_open(txtname, FST_TEXT);
  if (txtin == NULL) { die("unable to load text '%s'\n", txtname); }
  debug("%d sentence(s) in %s\n", txtin->nbelems, txtname);
    
  autalmot_t * pattern = load_grammar_automaton(patternname);

  if (pattern == NULL) { die("unable to load '%s' automaton\n", patternname); }

  autalmot_t * A;
  int no = 0;
  int totalmatches = 0;
  
  while ((A = fst_file_autalmot_load_next(txtin)) != NULL) {
    
    printf("sentence %d:\n", no + 1);
    if (A->nbstates < 2) {
      error("sentence %d is void\n", no + 1);
    } else {

      add_limphrase(A); 
      totalmatches = totalmatches + autalmot_locate_pattern(A, pattern, stdout);
    }
    autalmot_delete(A);
    no++;
  }

  printf("%d sentences proceed -- %d matching sequences.\n", no, totalmatches);

  return 0;
}

