 /*
  * Unitex
  *
  * Copyright (C) 2001-2006 Universit de Marne-la-Valle <unitex@univ-mlv.fr>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  * 
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
  *
  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Copyright.h"
#include "unicode.h"
#include "Arbre_dico.h"
#include "String_hash.h"
#include "Minimize_tree.h"
#include "Arbre_to_bin.h"
#include "DELA.h"
#include "IOBuffer.h"


#define MAX_TOKEN_LENGTH 1024
#define SEUIL_DE_DECISION 3
#define MAX_FREQUENT_WORDS 40
#define ERROR_MESSAGE "Cannot identify text language"

char lettres[0xFFFF];


struct cell {
    int N;
    int N2;
    unichar nom[256];
    struct cell* suivant;
};


struct arbre {
	int controle;
	struct transition* trans;
};

struct transition {
	unichar car;
	struct arbre* noeud;
	struct transition* suivant;
};


struct cell* creer_cell(int n,int n2,char* s,struct cell* suivant) {
struct cell* C=(struct cell*)malloc(sizeof(struct cell));
C->N=n;
C->N2=n2;
u_strcpy_char(C->nom,s);
C->suivant=suivant;
return C;
}

struct cell* creer_cell(int n,unichar* s,struct cell* suivant) {
struct cell* C=(struct cell*)malloc(sizeof(struct cell));
C->N=n;
C->N2=n;
u_strcpy(C->nom,s);
C->suivant=suivant;
return C;
}

void liberer_cell_liste(struct cell* l) {
struct cell* tmp;
while (l!=NULL) {
   tmp=l->suivant;
   free(l);
   l=tmp;
}
}


struct cell* inserer_dans_liste_triee(struct cell* l,int n,int n2,char* s) {
if (l==NULL) {
   return creer_cell(n,n2,s,NULL);
}
if (n>l->N) {
   l->suivant=inserer_dans_liste_triee(l->suivant,n,n2,s); 
   return l;
}
return creer_cell(n,n2,s,l);
}


struct cell* inserer_dans_liste_triee_inverse(struct cell* l,int n,unichar* s) {
if (l==NULL) {
   return creer_cell(n,s,NULL);
}
if (n<l->N) {
   l->suivant=inserer_dans_liste_triee_inverse(l->suivant,n,s); 
   return l;
}
return creer_cell(n,s,l);
}


//
// Par dfaut, on est trs permissif sur ce qui peut tre une lettre
//
void init_lettres() {
for (int i=0;i<=0x20;i++) {lettres[i]=0;}
for (int i=0x21;i<0xFFFF;i++) {lettres[i]=1;}
lettres['!']=0;
lettres['"']=0;
lettres['#']=0;
lettres['$']=0;
lettres['%']=0;
lettres['&']=0;
lettres['(']=0;
lettres[')']=0;
lettres['*']=0;
lettres['+']=0;
lettres[',']=0;
lettres['-']=0;
lettres['.']=0;
lettres['/']=0;
for (int i='0';i<='9';i++) {lettres[i]=0;}
lettres[':']=0;
lettres[';']=0;
lettres['<']=0;
lettres['=']=0;
lettres['>']=0;
lettres['?']=0;
lettres['@']=0;
lettres['[']=0;
lettres['\\']=0;
lettres[']']=0;
lettres['^']=0;
lettres['_']=0;
lettres['[']=0;
lettres['{']=0;
lettres['|']=0;
lettres['}']=0;
lettres['~']=0;
lettres[0x7F]=0;
}


char est_lettre(unichar c) {
return lettres[c];
}



char lire_mot(FILE* f,unichar* mot) {
	if (f==NULL) return 0;
	int c;
	while ((c=u_fgetc(f))!=EOF && !est_lettre(c));
	if (c==EOF) return 0;
	mot[0]=(unichar)c;
	int i=1;
	while ((i<(MAX_TOKEN_LENGTH-1)) && (c=u_fgetc(f))!=EOF && est_lettre(c)) {
		mot[i++]=(unichar)c;
	}
	mot[i]='\0';
	return 1;
}





struct arbre* creer_arbre() {
	struct arbre* a=(struct arbre*)malloc(sizeof(struct arbre));
	a->controle=0;
	a->trans=NULL;
	return a;
}


struct transition* creer_transition() {
	struct transition* t=(struct transition*)malloc(sizeof(struct transition));
    t->car='\0';
	t->noeud=NULL;
	t->suivant=NULL;
	return t;
}




struct arbre* ins(struct arbre* Arbre,unichar* mot,int pos) {
	if (Arbre==NULL) {
		Arbre=creer_arbre();
	}
	if (mot[pos]=='\0') {
		Arbre->controle=1;
		return Arbre;
	}
	struct transition* tmp=Arbre->trans;
	while (tmp!=NULL && tmp->car!=mot[pos]) {
		tmp=tmp->suivant;
	}
	if (tmp==NULL) {
		tmp=creer_transition();
		tmp->car=mot[pos];
		tmp->noeud=NULL;
		tmp->suivant=Arbre->trans;
		Arbre->trans=tmp;
	}
	tmp->noeud=ins(tmp->noeud,mot,pos+1);
	return Arbre;
}


struct arbre* inserer_dans_arbre(struct arbre* a,unichar* mon_mot) {
	return ins(a,mon_mot,0);
}


char existe(struct arbre* a,unichar* mot,int pos) {
	if (a==NULL) return 0;
	if (mot[pos]=='\0') return 1;
	struct transition* t=a->trans;
	while(t!=NULL) {
		if (mot[pos]==t->car) {
			return existe(t->noeud,mot,pos+1);
		}
		else t=t->suivant;
	}
	return 0;
}



char est_dans_l_arbre(struct arbre* a,unichar* mot) {
	return existe(a,mot,0);
}




void aff(struct arbre* a,unichar* tmp,int pos) {
	if (a==NULL) return;
	tmp[pos]='\0';
	u_prints(tmp); printf("\n");
	struct transition* t=a->trans;
	while(t!=NULL) {
		tmp[pos]=t->car;
		aff(t->noeud,tmp,pos+1);
		t=t->suivant;
	}
}


#define MAX_TOKENS 100000
int n_occur[MAX_TOKENS];

struct arbre* charger_liste_mots(char* s,char* frequent_words) {
    struct arbre* Arbre=NULL;
	FILE* f=u_fopen(s,U_READ);
	if (f==NULL) {
	   fprintf(stderr,"Cannot open %s\n",s);
	   return NULL;
	}
	FILE* freq=u_fopen(frequent_words,U_WRITE);
	if (freq==NULL) {
	   fprintf(stderr,"Cannot open %s\n",frequent_words);
	   u_fclose(f);
	   return NULL;
	}
	struct string_hash* hash=new_string_hash();
	for (int i=0;i<MAX_TOKENS;i++) {
	   n_occur[i]=0;
    }
	unichar tmp[MAX_TOKEN_LENGTH];
	while(lire_mot(f,tmp)) {
	    int n=get_hash_number(tmp,hash);
	    n_occur[n]++;
		int l=u_strlen(tmp);
		for (int i=0;i<l;i++) {
			Arbre=inserer_dans_arbre(Arbre,&(tmp[i]));
		}
	}
	u_fclose(f);
	if (Arbre==NULL) {
	    fprintf(stderr,"No token found in %s\n",s);
	}
	struct cell* liste=NULL;
	for (int i=0;i<hash->N;i++) {
	   liste=inserer_dans_liste_triee_inverse(liste,n_occur[i],hash->tab[i]);
	}
	free_string_hash(hash);
	struct cell* l=liste;
	int u=0;
	while (l!=NULL && u<MAX_FREQUENT_WORDS) {
	   u++;
	   u_fprints(l->nom,freq);
	   u_fprints_char("\n",freq);
	   l=l->suivant;
	}
	liberer_cell_liste(liste);
	u_fclose(freq);
	return Arbre;
}


void liberer_transition(struct transition*);

void liberer_arbre(struct arbre* a) {
	if (a==NULL) return;
	liberer_transition(a->trans);
	free(a);
}

void liberer_transition(struct transition* t) {
	struct transition* tmp;
	while(t!=NULL) {
		liberer_arbre(t->noeud);
		tmp=t->suivant;
		free(t);
		t=tmp;
	}
}



void explorer_arbre(struct arbre* racine,struct arbre* noeud,int longueur_maxi,
					unichar* tmp,int pos,char* alphabet,
                    struct arbre_dico* res,struct string_hash* hash,unichar* INFO) {
	if (pos>longueur_maxi) {
		// si on depasse la taille autorisee, on ne fait rien
		return;
	}
	// on va tester le mot qui est dans tmp
	// pour ca, on calcule les transitions qui n'existent pas
	char alph[0xFFFF];
	for (int i=0;i<0xFFFF;i++) {alph[i]=alphabet[i];}
	struct transition* t=noeud->trans;
	while (t!=NULL) {
		alph[t->car]=0;
		t=t->suivant;
	}
	for (int i=0;i<0xFFFF;i++) {
		if (alph[i]!=0) {
			// s'il n'y a pas de transition par alph[i]
			tmp[pos]=i;
			tmp[pos+1]='\0';
			if (est_dans_l_arbre(racine,&(tmp[1]))) {
				inserer_entree(tmp,INFO,res,hash);
			}
		}
	}
	t=noeud->trans;
	while (t!=NULL) {
		tmp[pos]=t->car;
		explorer_arbre(racine,t->noeud,longueur_maxi,tmp,pos+1,alphabet,res,hash,INFO);
		t=t->suivant;
	}
}



struct arbre_dico* calculer_facteurs_interdits(struct arbre* a,int longueur_maxi,char* alphabet) {
	struct arbre* racine=a;
	unichar tmp[MAX_TOKEN_LENGTH];
	// on construit l'alphabet autorise
	for (int i=0;i<0xFFFF;i++) {
		alphabet[i]=0;
	}
	struct transition* t=a->trans;
	while (t!=NULL) {
		alphabet[t->car]=1;
		t=t->suivant;
	}
	struct arbre_dico* res=new_arbre_dico();
	struct string_hash* hash=new_string_hash();
	tmp[1]='\0';
	unichar INFO[3];
	u_strcpy_char(INFO,".O");
	explorer_arbre(racine,a,longueur_maxi,tmp,0,alphabet,res,hash,INFO);
	/*for (int i=0;i<0xFFFF;i++) {
		if ((alphabet[i]==0) && u_is_letter((char)i)) {
			tmp[0]=i;
			inserer_entree(tmp,INFO,res,hash);
		}
	}*/
	return res;
}




int make_facteurs_interdits(char* nom,char* output,char* alph,char* freq) {
    printf("Loading reference corpus...\n");
	struct arbre* Arbre=charger_liste_mots(nom,freq);
	if (Arbre==NULL) {
	   return 1;
	}
	char alphabet[0xFFFF];
	printf("Computing forbidden factors...\n");
	struct arbre_dico* res=calculer_facteurs_interdits(Arbre,3,alphabet);
	FILE* ALPH=fopen(alph,"wb");
	if (fwrite(alphabet,1,0xFFFF,ALPH)!=(unsigned)0xFFFF) {
       fprintf(stderr,"Error while writing file %s\n",alph);
    }
    fclose(ALPH);
    liberer_arbre(Arbre);
	printf("Minimzing dictionary...\n");
	minimize_tree(res);
	printf("Saving...\n");
	creer_et_sauver_bin(res,output);
	printf("Done.\n");
	return 0;
}




char explorer_facteurs_interdits(unsigned char* BIN,int pos,unichar* contenu,unichar* facteur,
                                 int string_pos) {
int n_transitions;
n_transitions=((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1];
pos=pos+2;
if (!(n_transitions & 32768)) {
   // si on a trouve un facteur interdit minimal
   facteur[string_pos]='\0';
   // on est sur qu'il n'y en aura pas d'autre car les facteurs
   // sont minimaux
   return 1;
}
else {
    // si l'etat n'est pas minimal, on calcule le
    // nombre de transitions
    n_transitions=n_transitions-32768;
}
if (contenu[string_pos]=='\0') {
   // si on a fini de parcourir la chaine, on s'en va
   return 0;
}
for (int i=0;i<n_transitions;i++) {
  unichar c=(unichar)(((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1]);
  pos=pos+2;
  int adr=((unsigned char)BIN[pos])*256*256+((unsigned char)BIN[pos+1])*256+(unsigned char)BIN[pos+2];
  pos=pos+3;
  if (c==contenu[string_pos]) {
     facteur[string_pos]=c;
     if (explorer_facteurs_interdits(BIN,adr,contenu,facteur,string_pos+1)) {
          return 1;
     }
  }
}
return 0;
}


char match_facteur_interdit(unichar* mot,unsigned char* BIN,unichar* facteur) {
    return explorer_facteurs_interdits(BIN,4,mot,facteur,0);
}




int chercher_facteurs_interdits(char* entree,unsigned char* bin,char* alphabet,int *freq_words,struct string_hash* hash) {
FILE* f=u_fopen(entree,U_READ);
if (f==NULL) {
    fprintf(stderr,"Cannot open %s\n",entree);
    return -1;
}
int n=0;
unichar tmp[MAX_TOKEN_LENGTH];
unichar res[MAX_TOKEN_LENGTH];
while (lire_mot(f,tmp)) {
   if (get_token_number(tmp,hash)!=-1) {(*freq_words)++;}
   for (int i=0;i<u_strlen(tmp);i++) {
      if ((est_lettre(tmp[i]) && alphabet[tmp[i]]!=1) || match_facteur_interdit(&(tmp[i]),bin,res)) {
         n++;
      }
   }
}
u_fclose(f);
return n;
}


char lire_ligne(FILE *f,char* langue,char* fichier) {
int c;
int i=0;
while ((c=fgetc(f))!=EOF && c!='\n' && c!='=') {
   langue[i++]=(char)c;
}
if (c!='=') return 0;
langue[i]='\0';
i=0;
while ((c=fgetc(f))!=EOF && c!='\n') {
   fichier[i++]=(char)c;
}
fichier[i]='\0';
return 1;
}


void detecter_langue(char* liste,char* texte) {
struct cell* liste_triee=NULL;
unsigned char* bin;
char alphabet[0xFFFF];
FILE* LISTE=fopen(liste,"r");
if (LISTE==NULL) {
   fprintf(stderr,"Cannot open %s\n",liste);
   return;
}
int taille=u_file_size(texte);
char nom[1024];
char langue[64];
char langue_resultat[64];
int nbre_facteurs=-1;
int ecart=SEUIL_DE_DECISION+1;
char alph[1024];
char freq[1024];
struct string_hash* hash;
while (lire_ligne(LISTE,langue,nom)) {
   strcpy(alph,nom);
   strcat(alph,".alph");
   FILE* ALPH=fopen(alph,"rb");
   fread(alphabet,1,0xFFFF,ALPH);
   fclose(ALPH);
   strcpy(freq,nom);
   strcat(freq,".freq");
   hash=load_word_list(freq);
   bin=load_BIN_file(nom);
   if (bin==NULL) {
      fprintf(stderr,"Cannot open %s\n",nom);
   }
   else {
      int freq_words=0;
      int n=chercher_facteurs_interdits(texte,bin,alphabet,&freq_words,hash);
      //printf("%s: %d facteurs interdits   %d mots courants\n",langue,n,freq_words);
      liste_triee=inserer_dans_liste_triee(liste_triee,n,freq_words,langue);
      ecart=abs(nbre_facteurs-n);
      if (nbre_facteurs==-1 || n<nbre_facteurs) {
         nbre_facteurs=n;
         strcpy(langue_resultat,langue);
      }
      free(bin);
      free_string_hash(hash);
   }
}
fclose(LISTE);
if (liste_triee==NULL || nbre_facteurs==-1 || (nbre_facteurs>(taille/5) && (ecart<(taille/10))) || ecart<=SEUIL_DE_DECISION) {
   strcpy(langue_resultat,ERROR_MESSAGE);
}
else {
   //printf("Language= %s\n",langue_resultat); 
   //printf("Taille= %d\n",taille);
   //printf("Ecart=%d\n",ecart);
   //printf("-----------------------------\n");
   struct cell* C=liste_triee;
   int maximum=C->N+SEUIL_DE_DECISION;
   struct cell* liste2=NULL;
   struct cell* MAX=NULL;
   while (C!=NULL && C->N<=maximum) {
      if (MAX==NULL || MAX->N2<C->N2) {
         MAX=C;
      }
      C=C->suivant;
   }
   if (MAX->N2==0) {
      strcpy(langue_resultat,ERROR_MESSAGE);
   }
   else {
      u_to_char(langue_resultat,MAX->nom);
   }
   liberer_cell_liste(liste_triee);
   liberer_cell_liste(liste2);
}
printf("%s\n",langue_resultat);
}




void usage() {
printf("%s",COPYRIGHT);
printf("Usage:\n");
printf("  Lang -c <out> <corpus> : builds the forbidden factor dictionary\n");
printf("                           for the corpus <corpus>, and saves it in\n");
printf("                           the file named <out>\n");
printf("  Lang -l <list> <text> : detects the language of the file <text>\n");
printf("                          using the dictionaries listed in the file\n");
printf("                          <list>\n");
printf("\n");
}



int main(int argc,char** argv) {
setBufferMode();

    if (argc!=4) {
        usage();
        return 0;
    }
    
    init_lettres();
    
    if (!strcmp(argv[1],"-c")) {
        char tmp[1024];
        strcpy(tmp,argv[2]);
        strcat(tmp,".alph");
        char tmp2[1024];
        strcpy(tmp2,argv[2]);
        strcat(tmp2,".freq");
        make_facteurs_interdits(argv[3],argv[2],tmp,tmp2);
        return 0;
    }
    
    if (!strcmp(argv[1],"-l")) {
        detecter_langue(argv[2],argv[3]);
        return 0;
    }
    return 1;
}
