 /*
  * Unitex
  *
  * Copyright (C) 2001-2006 Universit de Marne-la-Valle <unitex@univ-mlv.fr>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  * 
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
  *
  */

//---------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "DELA.h"
#include "String_hash.h"
#include "unicode.h"
#include "Alphabet.h"
#include "Text_tokens.h"
#include "Dico_application.h"
#include "Liste_nombres.h"
#include "FileName.h"
#include "Token_tree.h"
#include "Table_complex_token_hash.h"
#include "Copyright.h"
#include "SpellChecking.h"
#include "IOBuffer.h"


//---------------------------------------------------------------------------


// 
// "C:\D\Seb\Unitex_1.2beta\French\Alphabet_sort.txt" "C:\D\Seb\Unitex_1.2beta\French\Dela\DELA.bin" "C:\D\scripts\Extraction\tempo_extract\err.txt" y "C:\D\scripts\Extraction\tempo_extract\result.txt"
//

//---------------------------------------------------------------------------
void usage() {
printf("%s",COPYRIGHT);
printf("Usage: SpellCheck <alphabet> <bin> <list> [yn] <out>\n");
printf("     <alphabet> : the alphabet file\n");
printf("     <bin> : the .bin file to use\n");
printf("     <list> : the list of words to be checked\n");
printf("     [yn] : indicates if the output concerns error words or not\n");
printf("     <out> : the list of output words\n\n");
printf("Filters words that are accent errors or errors dued to\n");
printf("a missing or extra letter.\n");
}



int main(int argc, char **argv) {
setBufferMode();

if (argc!=6) {
   usage();
   return 0;
}
unsigned char* bin;
if (!loadCharEquivalenceList(argv[1])) {
   fprintf(stderr,"Cannot load alphabet file %s\n",argv[1]);
   return 1;
}

bin=load_BIN_file(argv[2]);
if (bin==NULL) {
   fprintf(stderr,"Cannot open bin file %s\n",argv[2]);
   return 1;
}

FILE* f=u_fopen(argv[3],U_READ);
if (f==NULL) {
   free(bin);
   fprintf(stderr,"Cannot open file %s\n",argv[3]);
   return 1;
}

int KEEP_GOOD_WORDS;
if (!strcmp(argv[4],"y")) {
   KEEP_GOOD_WORDS=1;
}
else if (!strcmp(argv[4],"n")) {
        KEEP_GOOD_WORDS=0;
     }
     else {
        free(bin);
        u_fclose(f);
        fprintf(stderr,"Wrong parameter %s\n",argv[4]);
        return 1;
     }


FILE* output=u_fopen(argv[5],U_WRITE);
if (output==NULL) {
   free(bin);
   u_fclose(f);
   fprintf(stderr,"Cannot open output file %s\n",argv[5]);
   return 1;
}
unichar temp[1024];
unichar temp2[1024];

while (u_read_line(f,temp)) {
   int tmp=(!isSpellEquivalentWord(temp,bin)
       && !isSpellEquivalentWordModuloOneLetter(temp,bin,temp2))?1:0;
   if (KEEP_GOOD_WORDS==tmp) {
      u_fprints(temp,output);
      u_fprints_char("\n",output);
   }
}
free(bin);
u_fclose(f);
u_fclose(output);
return 0;
}
//---------------------------------------------------------------------------
