 /*
  * Unitex
  *
  * Copyright (C) 2001-2006 Universit de Marne-la-Valle <unitex@univ-mlv.fr>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  * 
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
  *
  */

#include "SpellCheckFrench.h"
#include "DELA.h"

int isCandidateForQuoteContraction(unichar*);
void buildQuoteCorrection(unichar*,unichar*);
int is_V_1s(dic_entry*);
int is_etre_V_3s(dic_entry*);
int is_avoir_V_3s(dic_entry*);
int is_V(dic_entry*);
int is_PPV(dic_entry*);


//
// Checks if 'w' can be viewed as a quote contraction error.
// Ex: "jaimerai" => "j'aimerai"
//
int isQuoteContraction(unichar* w,struct bin_dic* dic,
                       Alphabet* alph,
                       struct hypothesis_list** list,
                       int start_pos) {
if (w==NULL || dic==NULL) {
   return 0;
}
int pos=isCandidateForQuoteContraction(w);
if (!pos) {
   return 0;
}
int INF_number=get_INF_number(&(w[pos]),dic->bin,alph);
if (INF_number==NOT_IN_DICTIONARY) {
   return 0;
}
struct token_list* tmp=dic->inf->tab[INF_number];
unichar res[1024];
while (tmp!=NULL) {
   uncompress_entry(&(w[pos]),tmp->token,res);
   dic_entry* entry=tokenize_DELA_line(res);
   int OK=0;
   switch (w[0]) {
     case 'c':
     case 'C': {OK=is_etre_V_3s(entry) || is_avoir_V_3s(entry); break;}
     case 'd':
     case 'D': {OK=1; break;}
     case 'j':
     case 'J': {OK=is_V_1s(entry); break;}
     case 'l':
     case 'L': {OK=1; break;}
     case 'm':
     case 'M': {OK=is_V(entry) || is_PPV(entry); break;}
     case 'n':
     case 'N': {OK=is_V(entry) || is_PPV(entry); break;}
     case 'q':
     case 'Q': {OK=1; break;}
     case 's':
     case 'S': {OK=is_V(entry) || is_PPV(entry); break;}
     case 't':
     case 'T': {OK=is_V(entry) || is_PPV(entry); break;}
   }
   free_dic_entry(entry);
   if (OK) {
      buildQuoteCorrection(w,res);
      insert_in_hypothesis_list(list,
         new_spell_check_hypothesis(start_pos,
                                  start_pos+u_strlen(w)-1,
                                  SPELL_ERROR_SIMPLE_WORD,
                                  1,
                                  SPELL_QUOTE_ERROR,
                                  w,
                                  res,alph));
       return 1;
   }
   tmp=tmp->suivant;
}
return 0;
}




int isFrenchVowel(unichar c) {
return c=='a' || c=='A'
    || c==0xE0 || c==0xC0
    || c==0xE2 || c==0xC2
    || c==0xE4 || c==0xC4
    || c=='e' || c=='E'
    || c==0xE9 || c==0xC9
    || c==0xE8 || c==0xC8
    || c==0xEA || c==0xCA
    || c==0xEB || c==0xCB
    || c=='i' || c=='I'
    || c==0xEE || c==0xCE
    || c==0xEF || c==0xCF
    || c=='o' || c=='O'
    || c==0xF4 || c==0xD4
    || c==0xF6 || c==0xD6
    || c=='u' || c=='u'
    || c==0xF9 || c==0xD9
    || c==0xFB || c==0xDB
    || c==0xFC || c==0xDC
    || c=='y' || c=='Y'
    || c==0x153 || c==0x153
    || c=='h' || c=='H';
}



//
// Checks if 'w' is a valid candidate for a quote contraction
// error. 'w' must start by a letter of [c;d;j;l;m;n;s;t] followed
// by a vowel or an 'h'. Returns 0 if not. Otherwise, it returns
// the position of the hypothetical quote.
// Ex: "zone" => 0
//     "jaimerai" => 1
//     "quarrive" => 2
//
int isCandidateForQuoteContraction(unichar* w) {
int c=w[0];
if (c!='c' && c!='C'
 && c!='d' && c!='D'
 && c!='j' && c!='J'
 && c!='l' && c!='L'
 && c!='m' && c!='M'
 && c!='n' && c!='N'
 && c!='s' && c!='S'
 && c!='t' && c!='T'
 && ((c!='q' && c!='Q') || (w[1]!='u' && w[1]!='U'))) {
   return 0;
}
int p=(c=='q' || c=='Q')?2:1;
if (!isFrenchVowel(w[p])) {
  return 0;
}
return p;
}



//
// Creates from 'w' the quote correction.
// Ex: "Jaimerai" => "J'aimerai"
//
void buildQuoteCorrection(unichar* w,unichar* res) {
res[0]=w[0];
if (w[0]=='q' || w[0]=='Q') {
   res[1]=w[1];
   res[2]='\'';
   u_strcpy(&(res[3]),&(w[2]));
} else {
   res[1]='\'';
   u_strcpy(&(res[2]),&(w[1]));
}
}



//
// Checks if 'entry' conforms to V:1s
//
int is_V_1s(dic_entry* entry) {
unichar gram[2];
gram[0]='V';
gram[1]='\0';
unichar flex[3];
flex[0]='1';
flex[1]='s';
flex[2]='\0';
return dic_entry_contain_gram_code(entry,gram) &&
       dic_entry_contain_flex_code(entry,flex);
}



//
// Checks if 'entry' conforms to tre.V:3s
//
int is_etre_V_3s(dic_entry* entry) {
unichar etre[5];
etre[0]=0xEA;
etre[1]='t';
etre[2]='r';
etre[3]='e';
etre[4]='\0';
unichar gram[2];
gram[0]='V';
gram[1]='\0';
unichar flex[3];
flex[0]='3';
flex[1]='s';
flex[2]='\0';
return !u_strcmp(entry->lemma,etre) &&
       dic_entry_contain_gram_code(entry,gram) &&
       dic_entry_contain_flex_code(entry,flex);
}


//
// Checks if 'entry' is "aurait" or "et"
//
int is_avoir_V_3s(dic_entry* entry) {
unichar eut[4];
eut[0]='e';
eut[1]=0xFB;
eut[2]='t';
eut[3]='\0';
return !u_strcmp_char(entry->inflected,"aurait") ||
       !u_strcmp(entry->inflected,eut);
}



//
// Checks if 'entry' conforms to V
//
int is_V(dic_entry* entry) {
unichar gram[2];
gram[0]='V';
gram[1]='\0';
return dic_entry_contain_gram_code(entry,gram);
}


//
// Checks if 'entry' is "y" or "en"
//
int is_PPV(dic_entry* entry) {
return !u_strcmp_char(entry->lemma,"y") ||
       !u_strcmp_char(entry->lemma,"en");
}

