 /*
  * Unitex
  *
  * Copyright (C) 2001-2006 Universit de Marne-la-Valle <unitex@univ-mlv.fr>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  * 
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
  *
  */

#include "SpellCheckHypothesis.h"
#include "SpellChecking.h"
#include "Error.h"



int spell_equivalent_ignoring_case(unichar* a,unichar* b,Alphabet* alph) {

}


struct spell_check_hypothesis* new_spell_check_hypothesis(int start,int end,
                                  int mode,int distance,int trust_value,
                                  unichar* w,unichar* correction,Alphabet* alph) {
struct spell_check_hypothesis* res=(struct spell_check_hypothesis*)malloc(sizeof(struct spell_check_hypothesis));
if (res==NULL) {
   fprintf(stderr,"malloc error 1 in new_spell_check_hypothesis\n");
   fatal_error(1);
}
res->pos_start=start;
res->pos_end=end;
res->mode=mode;
res->distance=distance;
res->trust_value=trust_value;
res->weight=0;
if (w==NULL) {
   fprintf(stderr,"NULL word error in new_spell_check_hypothesis\n");
   fatal_error(1);
}
res->checked_sequence=u_strdup(w);
if (res->checked_sequence==NULL) {
   fprintf(stderr,"malloc error 2 in new_spell_check_hypothesis\n");
   fatal_error(1);
}
if (correction==NULL || is_equal_ignore_case_and_quotes(correction,w,alph)) {
   // if there is no correction, or a correction
   // identical to the word
   res->correction=NULL;
}
else {
   res->correction=u_strdup(correction);
   if (res->correction==NULL) {
      fprintf(stderr,"malloc error 3 in new_spell_check_hypothesis\n");
      fatal_error(1);
   }
}
return res;
}



void free_spell_check_hypothesis(struct spell_check_hypothesis* h) {
if (h==NULL) return;
if (h->checked_sequence!=NULL) free(h->checked_sequence);
if (h->correction!=NULL) free(h->correction);
free(h);
}




int equal(struct spell_check_hypothesis* h1,struct spell_check_hypothesis* h2) {
if ((h1==NULL && h2!=NULL) ||
    (h1!=NULL && h2==NULL)) return 0;
if ((h1->checked_sequence==NULL && h2->checked_sequence!=NULL) ||
    (h1->checked_sequence!=NULL && h2->checked_sequence==NULL)) return 0;
if (h1->checked_sequence!=NULL && u_strcmp(h1->checked_sequence,h2->checked_sequence)) return 0;
if ((h1->correction==NULL && h2->correction!=NULL) ||
    (h1->correction!=NULL && h2->correction==NULL)) return 0;
if (h1->correction!=NULL && u_strcmp(h1->correction,h2->correction)) return 0;
return (h1->pos_start==h2->pos_start)
    && (h1->pos_end==h2->pos_end);
    // we do not test the mode and the distance, because different corrections 
    // can overlap like accent errors and letter replacement
}



int is_in_hypothesis_list(struct hypothesis_list* list,
                           struct spell_check_hypothesis* hypothesis) {
if (hypothesis==NULL) {
   fprintf(stderr,"NULL hypothesis in is_in_hypothesis_list\n");
   fatal_error(1);
}
while (list!=NULL) {
   if (equal(list->hypothesis,hypothesis)) {
      // if the hypothesis is present, we update its distance and trust value
      if (hypothesis->distance<list->hypothesis->distance) {
         list->hypothesis->distance=hypothesis->distance;
         list->hypothesis->trust_value=hypothesis->trust_value;
      }
      else if (hypothesis->distance<list->hypothesis->distance) {
              // if there is already an hypothesis with the same distance
              // but a lesser trust value, we update it
              if (list->hypothesis->trust_value<hypothesis->trust_value) {
                 list->hypothesis->trust_value=hypothesis->trust_value;
              }
      }
      return 1;
   }
   list=list->next;
}
return 0;
}



struct hypothesis_list* new_hypothesis_list() {
struct hypothesis_list* res=(struct hypothesis_list*)malloc(sizeof(struct hypothesis_list));
if (res==NULL) {
   fprintf(stderr,"malloc error in new_hypothesis_list\n");
   fatal_error(1);
}
res->hypothesis=NULL;
res->next=NULL;
return res;
}


void free_hypothesis_list(struct hypothesis_list* list) {
struct hypothesis_list* tmp;
while (list!=NULL) {
   free_spell_check_hypothesis(list->hypothesis);
   tmp=list->next;
   free(list);
   list=tmp;
}
}



void concat_hypothesis_list(struct hypothesis_list** list1,
                            struct hypothesis_list* list2) {
if (*list1==NULL) {
   (*list1)=list2;
   return;
}
struct hypothesis_list* tmp=(*list1);
while (tmp->next!=NULL) {
   tmp=tmp->next;
}
tmp->next=list2;
}



void insert_in_hypothesis_list(struct hypothesis_list** list,
                           struct spell_check_hypothesis* hypothesis) {
if (!is_in_hypothesis_list(*list,hypothesis)) {
   struct hypothesis_list* tmp=new_hypothesis_list();
   tmp->hypothesis=hypothesis;
   tmp->next=(*list);
   (*list)=tmp;
} else {
  free_spell_check_hypothesis(hypothesis);
}
}




void display_hypothesis_list(struct hypothesis_list* list) {
struct spell_check_hypothesis* tmp;
while (list!=NULL) {
   tmp=list->hypothesis;
   printf("[%d;%d] w=\"",tmp->pos_start,tmp->pos_end);
   u_prints(tmp->checked_sequence);
   printf("\" cor=");
   if (tmp->correction!=NULL) {
      printf("\"");
      u_prints(tmp->correction);
      printf("\"");
   }
   else {
      printf("(none)");
   }
   printf(" mode=%d d=%d  t=%d  weight=%d\n",tmp->mode,tmp->distance,tmp->trust_value,tmp->weight);
   list=list->next;
}
}



//
// Returns -1 if a<b
//          0 if a=b
//          1 if a>b
//
int compare_hypothesis_by_distance_and_trust(struct spell_check_hypothesis* a,
                       struct spell_check_hypothesis* b) {
if (a==NULL || b==NULL) {
   fprintf(stderr,"NULL error in compare_hypothesis\n");
   fatal_error(1);
}
if (a->distance<b->distance) {return -1;}
if (a->distance>b->distance) {return 1;}
return (b->trust_value)-(a->trust_value);
}



//
// Returns -1 if a<b
//          0 if a=b
//          1 if a>b
//
int compare_hypothesis_by_trust_and_distance(struct spell_check_hypothesis* a,
                       struct spell_check_hypothesis* b) {
if (a==NULL || b==NULL) {
   fprintf(stderr,"NULL error in compare_hypothesis\n");
   fatal_error(1);
}
if (a->distance<b->distance) {
   if (b->trust_value > (a->trust_value+TRUST_VALUE_THRESHOLD)) {
      // case: distance 1 trust 160|distance 2 trust 190
      return 1;
   } else {return -1;}
}
if (a->distance>b->distance) {
   if (a->trust_value > (b->trust_value+TRUST_VALUE_THRESHOLD)) {
     // case: distance 2 trust 190 distance 1 trust 160
     return -1;
   } else {return 1;}
}
return (b->trust_value)-(a->trust_value);
}




//
// Inserts the hypothesis of 'h' in 'list' at the correct place and returns the
// resulting list
//
struct hypothesis_list* insert_hypothesis_in_correct_position(
                                  struct hypothesis_list* h,
                                  struct hypothesis_list* list,
                                  int sort_function(struct spell_check_hypothesis*,struct spell_check_hypothesis*)) {
if (list==NULL) {
   // if we have the empty list, we just
   // return the element 'h'
   h->next=NULL;
   return h;
}
if (sort_function(h->hypothesis,list->hypothesis)<=0) {
   // if we must insert 'h' here
   h->next=list;
   return h;
}
// if must insert 'h' in the rest of the list
list->next=insert_hypothesis_in_correct_position(h,list->next,sort_function);
return list;
}



//
// Sorts an hypothesis list according to a comparison function
//
void sort_hypothesis_list(struct hypothesis_list** list
                 ,int sort_function(struct spell_check_hypothesis*,struct spell_check_hypothesis*)) {
if (*list==NULL || (*list)->next==NULL) {
   // empty lists and 1-item lists do not need to be sorted
   return;
}
struct hypothesis_list* tmp=(*list)->next;
sort_hypothesis_list(&tmp,sort_function);
(*list)=insert_hypothesis_in_correct_position(*list,tmp,sort_function);
}




void keep_first_hypothesis_with_same_distance(struct hypothesis_list** list) {
if (*list==NULL) {
   return;
}
int i=1;
struct hypothesis_list* tmp=*list;
int distance=tmp->hypothesis->distance;
while (tmp->next!=NULL && i<SPELL_MAX_HYPOTHESIS
       && tmp->next->hypothesis->distance==distance) {
   tmp=tmp->next;
   i++;
}
free_hypothesis_list(tmp->next);
tmp->next=NULL;
}



void keep_first_hypothesis(struct hypothesis_list** list) {
if (*list==NULL) {
   return;
}
int i=1;
struct hypothesis_list* tmp=*list;
int distance=tmp->hypothesis->distance;
while (tmp->next!=NULL && i<SPELL_MAX_HYPOTHESIS) {
   tmp=tmp->next;
   i++;
}
free_hypothesis_list(tmp->next);
tmp->next=NULL;
}




/*struct hypothesis_list* filter_hypothesis_list_by_distance_threshold(struct hypothesis_list* list,
                                         int threshold) {
if (list==NULL) {
   return NULL;
}
if (list->hypothesis->distance>threshold) {
   // if we must remove this hypothesis
   struct hypothesis_list* tmp=list->next;
   list->next=NULL;
   free_hypothesis_list(list);
   return filter_hypothesis_list_by_distance_threshold(tmp,threshold);
}
list->next=filter_hypothesis_list_by_distance_threshold(list->next,threshold);
return list;
}*/


struct hypothesis_list* filter_list_by_pos_end(struct hypothesis_list* list,
                                         int pos_end) {
if (list==NULL) {
   return NULL;
}
if (list->hypothesis->pos_end<pos_end) {
   // if we must remove this hypothesis
   struct hypothesis_list* tmp=list->next;
   list->next=NULL;
   free_hypothesis_list(list);
   return filter_list_by_pos_end(tmp,pos_end);
}
list->next=filter_list_by_pos_end(list->next,pos_end);
return list;
}



//
// This functions takes a list of hypothesis sorted by relevancy, and
// gives a weight of SPELL_MAX_HYPOTHESIS to the first,
// SPELL_MAX_HYPOTHESIS-1 to the second, ...
//
void compute_weights(struct hypothesis_list** list) {
int weight=SPELL_MAX_HYPOTHESIS;
struct hypothesis_list* tmp=(*list);
while (tmp!=NULL) {
   tmp->hypothesis->weight=weight;
   weight--;
   if (weight<0) {
      fprintf(stderr,"Inconsistency in compute_weights\n");
      fatal_error(1);
   }
   tmp=tmp->next;
}
}


//
// keep better solutions of a simple word list
//
void clean_hypothesis_list(struct hypothesis_list** list) {
if (*list==NULL) {
   // if there is nothing to do
   return;
}
sort_hypothesis_list(list,compare_hypothesis_by_distance_and_trust);
//----first case: the first hypothesis has a too bad distance
/*int reasonable_error_threshold=SPELL_DEFAULT_DISTANCE
                               +u_strlen((*list)->hypothesis->checked_sequence)/5;
(*list)=filter_hypothesis_list_by_distance_threshold(*list,reasonable_error_threshold);
*/
(*list)=filter_hypothesis_list_by_distance(*list);
if (*list==NULL) {
   // if the list has been emptied
   return;
}
//-----we try to take the first SPELL_MAX_HYPOTHESIS---------- 
keep_first_hypothesis_with_same_distance(list);
compute_weights(list);
}



int get_longuest_compound(struct hypothesis_list* list) {
if (list==NULL) {
   return -1;
}
int res=list->hypothesis->pos_end;
list=list->next;
while (list!=NULL) {
   if (list->hypothesis->pos_end>res) {
      res=list->hypothesis->pos_end;
   }
   list=list->next;
}
return res;
}




int starts_by_a_stop_word(unichar* w,struct string_hash* stop_words,Alphabet* alph) {
if (w==NULL) {
   fprintf(stderr,"NULL error in ends_by_a_stop_word\n");
   fatal_error(1);
}
unichar tmp[1024];
int i;
for (i=0;is_letter(w[i],alph);i++) {tmp[i]=w[i];}
tmp[i]='\0';
return is_in_string_hash_modulo_case(tmp,stop_words,alph);
}




int ends_by_a_stop_word(unichar* w,struct string_hash* stop_words,Alphabet* alph) {
if (w==NULL) {
   fprintf(stderr,"NULL error in ends_by_a_stop_word\n");
   fatal_error(1);
}
int pos=1+index_of_last_non_letter(w,alph);
return is_in_string_hash_modulo_case(&(w[pos]),stop_words,alph);
}




//
// returns a list where there is no hypothesis with a checked sequence
// ending with a stop word.
// Ex: "descending of"
//
struct hypothesis_list* remove_suffix_errors(struct hypothesis_list* list,
                                    struct string_hash* stop_words,
                                    Alphabet* alph) {
if (list==NULL) {
   return NULL;
}
if (list->hypothesis->mode!=SPELL_CORRECT_COMPOUND_WORD && 
    ends_by_a_stop_word(list->hypothesis->checked_sequence,stop_words,alph)) {
   // if we must remove this hypothesis
   struct hypothesis_list* tmp=list->next;
   list->next=NULL;
   free_hypothesis_list(list);
   return remove_suffix_errors(tmp,stop_words,alph);
}
list->next=remove_suffix_errors(list->next,stop_words,alph);
return list;
}



//
// This function removes hypothesis with a too big distance
//
struct hypothesis_list* filter_hypothesis_list_by_distance(struct hypothesis_list* list) {
if (list==NULL) {
   return NULL;
}
int d=(1+u_strlen(list->hypothesis->checked_sequence)/DISTANCE_COEFFICIENT);
if (d>DISTANCE_LIMIT) {d=DISTANCE_LIMIT;}
if (list->hypothesis->distance>d) {
   // if we must remove this hypothesis
   struct hypothesis_list* tmp=list->next;
   list->next=NULL;
   free_hypothesis_list(list);
   return filter_hypothesis_list_by_distance(tmp);
}
list->next=filter_hypothesis_list_by_distance(list->next);
return list;
}



//
// keep better solutions of a compound word list
//
void clean_compound_hypothesis_list(struct hypothesis_list** list,
                                    struct string_hash* stop_words,
                                    Alphabet* alph) {
if (*list==NULL) {
   // if there is nothing to do
   return;
}
(*list)=remove_suffix_errors(*list,stop_words,alph);
(*list)=filter_hypothesis_list_by_distance(*list);
//int max=get_longuest_compound(*list);
//(*list)=filter_list_by_pos_end(*list,max);
clean_hypothesis_list(list);
}



void compute_distance_and_trust_value(char* operations,int *distance,
                                      int *trust_value) {
char tmp[1024];
clear_operations(operations,tmp);
(*distance)=strlen(tmp);
(*trust_value)=0;
for (int i=0;i<tmp[i]!='\0';i++) {
   switch (tmp[i]) {
      case OP_INSERTION: (*trust_value)+=SPELL_BONUS_INSERTION; break;
      case OP_DELETION: (*trust_value)+=SPELL_BONUS_DELETION; break;
      case OP_DUPLICATION: (*trust_value)+=SPELL_BONUS_DUPLICATION; break;
      case OP_SWAP: (*trust_value)+=SPELL_BONUS_LETTER_INVERSION; break;
      case OP_REPLACEMENT: (*trust_value)+=SPELL_BONUS_LETTER_REPLACEMENT; break;
      case OP_REPLACEMENT_ACCENT: (*trust_value)+=SPELL_BONUS_ACCENT; break;
      case OP_REPLACEMENT_UPPERCASE: (*trust_value)+=SPELL_BONUS_UPPERCASE; break;
      case OP_REPLACEMENT_CLOSE_KEY: (*trust_value)+=SPELL_BONUS_CLOSE_ON_KEYBOARD; break;
   }
}
}



//
// this function takes a string representing an operation sequence,
// and cleans it as follows:
// 1) when there is a swap tag, remove the operation sequence that is represented
//    by this tag. Ex: "RRX" => "X"
// 2) remove normal operations. Ex: " D I  R" => "DIR"
//
void clear_operations(char* operations,char* dst) {
int j=0;
for (int i=0;operations[i]!='\0';i++) {
   switch (operations[i]) {
      case OP_NORMAL:break;
      case OP_SWAP: dst[j-2]=OP_SWAP; j=j-1; break;
      default: dst[j]=operations[i]; j++;
   }
}
dst[j]='\0';
}




int compare_word_errors(struct spell_check_hypothesis* a,
                       struct spell_check_hypothesis* b) {
if (a==NULL || b==NULL) {
   fprintf(stderr,"NULL error in compare_simple_word_errors\n");
   fatal_error(1);
}
if (a->mode==SPELL_ERROR_SIMPLE_WORD ||
    a->mode==SPELL_ERROR_COMPOUND_WORD ||
    a->mode==SPELL_WRONG_WORD_WITH_NO_HYPOTHESIS) {return -1;}
if (b->mode==SPELL_ERROR_SIMPLE_WORD ||
    b->mode==SPELL_ERROR_COMPOUND_WORD ||
    b->mode==SPELL_WRONG_WORD_WITH_NO_HYPOTHESIS) {return 1;}
return 0;
}

//
// This function sorts the list by placing simple word errors
// at the beginning of it
//
void put_errors_at_beginning(struct hypothesis_list** list) {
sort_hypothesis_list(list,compare_word_errors);
}



int is_included_hypothesis_in_correct_compound_word(
                  struct spell_check_hypothesis* hypothesis,
                  struct hypothesis_list* list) {
if (hypothesis==NULL) {
   fprintf(stderr,"NULL error in is_included_hypothesis_in_correct_compound_word\n");
   fatal_error(1);
}
while (list!=NULL) {
   if (list->hypothesis->mode==SPELL_CORRECT_COMPOUND_WORD) {
      // if the hypothesis is a compound word, we check if it includes
      // 'hypothesis'
      if (hypothesis->pos_start >= list->hypothesis->pos_start
          && hypothesis->pos_end <= list->hypothesis->pos_end) {
         return 1;
      }
   }
   list=list->next;
}
return 0;
}



//
// This function tests if the hypothesis of 'list' is a simple word error.
// If not, we return because the list is supposed to be sorted with simple 
// word errors at the beginning. Otherwise, it checks if this hypothesis
// is included in a correct compound word, and in that case, it discard the
// hypothesis
//
struct hypothesis_list* remove_false_simple_or_compound_word_errors(struct hypothesis_list* list) {
if (list==NULL) {
   return NULL;
}
if (list->hypothesis->mode!=SPELL_ERROR_SIMPLE_WORD &&
    list->hypothesis->mode!=SPELL_ERROR_COMPOUND_WORD &&
    list->hypothesis->mode!=SPELL_WRONG_WORD_WITH_NO_HYPOTHESIS) {
   return list;
}
if (is_included_hypothesis_in_correct_compound_word(list->hypothesis,list->next)) {
   struct hypothesis_list* tmp=list->next;
   list->next=NULL;
   free_hypothesis_list(list);
   return remove_false_simple_or_compound_word_errors(tmp);
}
list->next=remove_false_simple_or_compound_word_errors(list->next);
return list;
}


//
// This function discards false simple words errors, like
// "aujourd" in "aujourd'hui"
//
void filter_unkown_part_of_compound_words(struct hypothesis_list** list) {
put_errors_at_beginning(list);
(*list)=remove_false_simple_or_compound_word_errors(*list);
}



struct hypothesis_list* remove_correct_compound_words(
                                   struct hypothesis_list* list,
                                   int pos) {
if (list==NULL) {
   return NULL;
}
if (list->hypothesis->mode==SPELL_CORRECT_COMPOUND_WORD
    && list->hypothesis->pos_end < pos) {
   // if we must remove the hypothesis
   struct hypothesis_list* tmp=list->next;
   list->next=NULL;
   free_hypothesis_list(list);
   return remove_correct_compound_words(tmp,pos);
}
list->next=remove_correct_compound_words(list->next,pos);
return list;
}

//
// Removes from the list every compound with an end position < pos
//
void remove_correct_compound_words(struct hypothesis_list** list,int pos) {
(*list)=remove_correct_compound_words(*list,pos);
}


struct hypothesis_list* remove_unkown_simple_word_with_no_suggestion (
                                   struct hypothesis_list* list) {
if (list==NULL) {
   return NULL;
}
if (list->hypothesis->mode==SPELL_WRONG_WORD_WITH_NO_HYPOTHESIS) {
   // if we must remove the hypothesis
   struct hypothesis_list* tmp=list->next;
   list->next=NULL;
   free_hypothesis_list(list);
   return remove_unkown_simple_word_with_no_suggestion(tmp);
}
list->next=remove_unkown_simple_word_with_no_suggestion(list->next);
return list;
}

//
// Removes from the list every unknown word with no hypothesis
//
void remove_unkown_simple_words_with_no_suggestion(struct hypothesis_list** list) {
(*list)=remove_unkown_simple_word_with_no_suggestion(*list);
}





int compare_by_length(struct spell_check_hypothesis* a,
                       struct spell_check_hypothesis* b) {
if (a==NULL || b==NULL) {
   fprintf(stderr,"NULL error in compare_simple_word_errors\n");
   fatal_error(1);
}
int length_a=a->pos_end-a->pos_start;
int length_b=b->pos_end-b->pos_start;
return length_a-length_b;
}



int compare_by_length_descending(struct spell_check_hypothesis* a,
                                 struct spell_check_hypothesis* b) {
if (a==NULL || b==NULL) {
   fprintf(stderr,"NULL error in compare_simple_word_errors\n");
   fatal_error(1);
}
int length_a=a->pos_end-a->pos_start;
int length_b=b->pos_end-b->pos_start;
return length_b-length_a;
}




//
// Sorts the list by ascending length of checked_sequence
//
void sort_hypothesis_by_length(struct hypothesis_list** list) {
sort_hypothesis_list(list,compare_by_length);
}



//
// Sorts the list by descending length of checked_sequence
//
void sort_hypothesis_by_length_descending(struct hypothesis_list** list) {
sort_hypothesis_list(list,compare_by_length_descending);
}



//$$$$$$$$$$$$$$$$$ BEGIN FINAL CLEAN $$$$$$$$$$$$$$$$$$$$$$$



//
// Returns -1 if a<b
//          0 if a=b
//          1 if a>b
//
int compare_hypothesis_by_distance_weight_and_mode(struct spell_check_hypothesis* a,
                       struct spell_check_hypothesis* b) {
if (a==NULL || b==NULL) {
   fprintf(stderr,"NULL error in compare_hypothesis\n");
   fatal_error(1);
}
if (a->distance<b->distance) {return -1;}
if (a->distance>b->distance) {return 1;}
int w=(b->weight)-(a->weight);
if (w) return w;
if (a->mode==SPELL_ERROR_COMPOUND_WORD) {
   if (a->mode==SPELL_ERROR_COMPOUND_WORD) {
      return 0;
   } else {
     return -1;
   }
}
if (b->mode==SPELL_ERROR_COMPOUND_WORD) {
   return 1;
}
return 0;
}




void keep_first_trust_value_island(struct hypothesis_list** list) {
if (*list==NULL) {
   return;
}
struct hypothesis_list* tmp=*list;


while (tmp->next!=NULL && (tmp->hypothesis->trust_value<=tmp->next->hypothesis->trust_value+TRUST_VALUE_ISLAND_THRESHOLD)) {
   tmp=tmp->next;
}
free_hypothesis_list(tmp->next);
tmp->next=NULL;
}




//
// Keeps better solutions of a list. This function is invoked just
// before writing hypothesis to the output
//
void final_clean_hypothesis_list(struct hypothesis_list** list) {
if (*list==NULL) {
   // if there is nothing to do
   return;
}
sort_hypothesis_list(list,compare_hypothesis_by_distance_weight_and_mode);
//----first case: the first hypothesis has a too bad distance
if (*list==NULL) {
   // if the list has been emptied
   return;
}
//-----we try to take the first SPELL_MAX_HYPOTHESIS---------- 
keep_first_hypothesis(list);
keep_first_trust_value_island(list);
}

//$$$$$$$$$$$$$$$$$ END FINAL CLEAN $$$$$$$$$$$$$$$$$$$$$$$


