 /*
  * Unitex
  *
  * Copyright (C) 2001-2006 Universit de Marne-la-Valle <unitex@univ-mlv.fr>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  * 
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
  *
  */
  
#ifndef SpellCheckHypothesisH
#define SpellCheckHypothesisH

#include "unicode.h"
#include "Alphabet.h"

#define SPELL_CORRECT_SIMPLE_WORD 0
#define SPELL_CORRECT_COMPOUND_WORD 1
#define SPELL_ERROR_COMPOUND_WORD 2
#define SPELL_ERROR_SIMPLE_WORD 3
#define SPELL_WRONG_WORD_WITH_NO_HYPOTHESIS 4
#define SPELL_OVERLAP_RESULT 5


// for a word of length L, we will tolerate
// a distance of 1+L/DISTANCE_COEFFICIENT, in
// a limit of DISTANCE_LIMIT
#define DISTANCE_COEFFICIENT 7
#define DISTANCE_LIMIT 2

// if all hypothesis are compounds, then they must not have
// a distance greater than DISTANCE_MAX_FOR_ONLY_COMPOUNDS
#define DISTANCE_MAX_FOR_ONLY_COMPOUNDS 1

// words with length<this limit won't be taken
// into account for approximate matching in order
// to avoid noise
//#define SPELL_MINIMUM_SIZE_FOR_APPROXIMATE_MATCHING 4

// default maximum distance for approximate matching
#define SPELL_DEFAULT_DISTANCE_FOR_SIMPLE_WORDS 3
#define SPELL_DEFAULT_DISTANCE_FOR_COMPOUND_WORDS 1


#define SPELL_EMPTY_TRUST_VALUE 0


//------------------------------------
#define SPELL_BONUS_DUPLICATION 25
#define SPELL_BONUS_UPPERCASE 15
#define SPELL_BONUS_ACCENT 30
#define SPELL_BONUS_CLOSE_ON_KEYBOARD 15
#define SPELL_BONUS_LETTER_REPLACEMENT 10
#define SPELL_BONUS_LETTER_INVERSION 30
#define SPELL_BONUS_INSERTION 10
#define SPELL_BONUS_DELETION 10
#define SPELL_BONUS_COMPOUND_SEPARATOR_ERROR 30
#define SPELL_QUOTE_ERROR 50
//------------------------------------

//maximum number of hypothesis that will be keeped for each sequence
#define SPELL_MAX_HYPOTHESIS 6

// used to sort cases like "distance 1 trust 160|distance 2 trust 190"
// according to the trust_value, if there is a difference of more
// than TRUST_VALUE_THRESHOLD
#define TRUST_VALUE_THRESHOLD 20


// used to remove hypothesis, when there is a better one with
// a trust_value greater of at least TRUST_VALUE_ISLAND_THRESHOLD
// Ex: A=distance 2,trust_value 50; B=distance 2,trust_value 45;
//     C=distance 2,trust_value 20
//
// => we only keep A and B, because the TRUST_VALUE_ISLAND_THRESHOLD
// is overpassed from B to C
#define TRUST_VALUE_ISLAND_THRESHOLD 20


//-----------------------------------------------------------------------------
#define OP_NORMAL ' '                   // "Pierre" => "pierre"
#define OP_INSERTION 'I'                // "cavier" => "clavier"
#define OP_DELETION 'X'                 // "cklavier" => "clavier"
#define OP_DUPLICATION 'D'              // "pome" => "pomme"
#define OP_SWAP 'S'                     // "lcavier" => "clavier"
#define OP_REPLACEMENT 'R'              // "czavier" => "clavier"
#define OP_REPLACEMENT_ACCENT 'A'       // "mangees" => "manges"
#define OP_REPLACEMENT_UPPERCASE 'U'    // "Linus torvald" => "Linus Torvald"
#define OP_REPLACEMENT_CLOSE_KEY 'K'    // "cmavier" => "clavier"
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
#define ALL_IN_UPPER_CASE_DISTANCE 1
#define ALL_IN_UPPER_CASE_TRUST_VALUE 100
//-----------------------------------------------------------------------------



struct spell_check_hypothesis {
   int pos_start;
   int pos_end;
   int mode;
   int distance;
   int trust_value;
   int weight;
   unichar* checked_sequence;
   unichar* correction;
};


struct hypothesis_list {
   struct spell_check_hypothesis* hypothesis;
   struct hypothesis_list* next;
};



struct spell_check_hypothesis* new_spell_check_hypothesis(int,int,int,int,int,unichar*,unichar*,Alphabet*);
void free_spell_check_hypothesis(struct spell_check_hypothesis*);
int is_in_hypothesis_list(struct hypothesis_list*,struct spell_check_hypothesis*);
void insert_in_hypothesis_list(struct hypothesis_list**,struct spell_check_hypothesis*);
void free_hypothesis_list(struct hypothesis_list*);
void display_hypothesis_list(struct hypothesis_list*);

int compare_hypothesis_by_distance_and_trust(struct spell_check_hypothesis*,
                       struct spell_check_hypothesis*);
int compare_hypothesis_by_trust_and_distance(struct spell_check_hypothesis*,
                       struct spell_check_hypothesis*);
void sort_hypothesis_list(struct hypothesis_list**,int (struct spell_check_hypothesis*,struct spell_check_hypothesis*));
void clean_hypothesis_list(struct hypothesis_list**);

void compute_distance_and_trust_value(char*,int*,int*);

void clear_operations(char*,char*);

void clean_compound_hypothesis_list(struct hypothesis_list**,struct string_hash*,Alphabet*);

int starts_by_a_stop_word(unichar*,struct string_hash*,Alphabet*);
int ends_by_a_stop_word(unichar*,struct string_hash*,Alphabet*);

void concat_hypothesis_list(struct hypothesis_list**,struct hypothesis_list*);void concat_hypothesis_list(struct hypothesis_list** list1,
                            struct hypothesis_list* list2);

void filter_unkown_part_of_compound_words(struct hypothesis_list**);
void remove_correct_compound_words(struct hypothesis_list**,int);

void sort_hypothesis_by_length(struct hypothesis_list**);
void sort_hypothesis_by_length_descending(struct hypothesis_list**);
void final_clean_hypothesis_list(struct hypothesis_list**);
struct hypothesis_list* filter_hypothesis_list_by_distance(struct hypothesis_list*);
void remove_unkown_simple_words_with_no_suggestion(struct hypothesis_list**);
#endif

