 /*
  * Unitex
  *
  * Copyright (C) 2001-2006 Universit de Marne-la-Valle <unitex@univ-mlv.fr>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  * 
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
  *
  */

#include "SpellChecking.h"
#include "Liste_nombres.h"
#include "unicode.h"
#include "SpellCheckHypothesis.h"
#include "Error.h"
#include "SpellCheckFrench.h"


static int tab[0xFFFF];
static struct liste_nombres* tab2[0xFFFF];
static int N;

void initTab() {
for (int i=0;i<0xFFFF;i++) {
   tab[i]=-1;
   tab2[i]=NULL;
}
N=0;
}


int loadCharEquivalenceList(char* n) {
FILE* f=u_fopen(n,U_READ);
if (f==NULL) return 0;
initTab();
int c;
do {
   while ((c=u_fgetc(f))!='\n' && c!=EOF) {
      tab[c]=N;
      tab2[N]=inserer_dans_liste_nombres(c,tab2[N]);
   }
   N++;
} while (c!=EOF);
u_fclose(f);
return 1;
}



int areSpellEquivalentLetters(unichar a,unichar b) {
if (tab[a]==-1) return 0;
struct liste_nombres* l=tab2[tab[a]];
while (l!=NULL) {
   if (l->n==b) return 1;
   l=l->suivant;
}
return 0;
}




//
// Returns 1 if 'c' can be removed in a correction.
//
int is_removable_char(unichar c) {
return c!=' ' && c!='\'' && c!='\0' && (c<'0' || c>'9')
              && c!='.' && c!=',' && c!=';' && c!='"'
              && c!='-' && c!='(' && c!=')';
}



//
// Returns if the chars are both letters or both digits
//
int can_replace_chars(unichar original,unichar correction,Alphabet* alph) {
if (is_letter(original,alph)) {
   return is_letter(correction,alph);
}
if (original>'0' && original<'9') {
   return (correction>'0' && correction<'9');
}
return 0;
}



int exploreBin(int pos,unichar* word,int string_pos,unsigned char* BIN) {
int n_transitions;
n_transitions=((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1];
pos=pos+2;
if (word[string_pos]=='\0') {
   // if we are at the end of the word, we just check
   // if the state is final, i.e. if the word belongs
   // to the dictionary
   return (!(n_transitions & 32768));
}
if ((n_transitions & 32768)) {
   // if we are in a normal node, we remove the control bit to
   // have the good number of transitions
   n_transitions=n_transitions-32768;
} else {
  // if we are in a final node, we must jump after the reference to the INF line number
  pos=pos+3;
}
for (int i=0;i<n_transitions;i++) {
  unichar c=(unichar)(((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1]);
  pos=pos+2;
  int adr=((unsigned char)BIN[pos])*256*256+((unsigned char)BIN[pos+1])*256+(unsigned char)BIN[pos+2];
  pos=pos+3;
  if (areSpellEquivalentLetters(c,word[string_pos])
      || are_equivalent_quotes(c,word[string_pos])) {
     if (exploreBin(adr,word,string_pos+1,BIN)) return 1;
  }
}
return 0;
}


int exploreBin2(int pos,unichar* word,int string_pos,unsigned char* BIN,int nErrors,
                unichar* w,int w_pos) {
int n_transitions;
// we explore removing a letter: abc -> bc
if (nErrors>0) {
   if (exploreBin2(pos,word,string_pos+1,BIN,nErrors-1,w,w_pos)) return 1;
}

n_transitions=((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1];
pos=pos+2;
if (word[string_pos]=='\0') {
   // if we are at the end of the word, we just check
   // if the state is final, i.e. if the word belongs
   // to the dictionary
   w[w_pos]='\0';
   return (nErrors==0 && !(n_transitions & 32768));
}
if ((n_transitions & 32768)) {
   // if we are in a normal node, we remove the control bit to
   // have the good number of transitions
   n_transitions=n_transitions-32768;
} else {
  // if we are in a final node, we must jump after the reference to the INF line number
  pos=pos+3;
}
for (int i=0;i<n_transitions;i++) {
  unichar c=(unichar)(((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1]);
  w[w_pos]=c;
  pos=pos+2;
  int adr=((unsigned char)BIN[pos])*256*256+((unsigned char)BIN[pos+1])*256+(unsigned char)BIN[pos+2];
  pos=pos+3;
  // we explore normally: abc -> abc
  if (c==word[string_pos]
      || are_equivalent_quotes(c,word[string_pos])) {
     if (exploreBin2(adr,word,string_pos+1,BIN,nErrors,w,w_pos+1)) return 1;
  }
  // we explore replacing a letter: abc -> zbc
  //if (nErrors>0 && c!=word[string_pos]) {
  //   if (exploreBin2(adr,word,string_pos+1,BIN,nErrors-1,w,w_pos+1)) return 1;
  //}
  // we explore duplicating a letter: abc -> aabc
  if (nErrors>0 && c==word[string_pos]) {
     if (exploreBin2(adr,word,string_pos,BIN,nErrors-1,w,w_pos+1)) return 1;
  }
}
return 0;
}


//
// this function tests in w is a word in bin modulo letter equivalency
// like accent or case variants
// ex: "london" and "Londn" tolerated for "London"
// 
int isSpellEquivalentWord(unichar* w,unsigned char* bin) {
if (w==NULL || bin==NULL) return -1;
return exploreBin(4,w,0,bin);
}



int isSpellEquivalentWordModuloOneLetter(unichar* w,unsigned char* bin,unichar* original) {
original[0]='\0';
if (w==NULL || bin==NULL) {
   return -1;
}
if (u_strlen(w)<5) return 0;
int res=exploreBin2(4,w,0,bin,1,original,0);
if (res!=1) {
   original[0]='\0';
}
return res;
}



//
// This function is called when 'dic_letter' replaces 'text_letter' in 
// a correction hypothesis. It computes a bonus that indicates if
// the two letters are close from several points of view.
//
int replacement_operation(unichar text_letter,unichar dic_letter,Alphabet* alph) {
if (text_letter==dic_letter) {
   // this case shouldn't happen
   fprintf(stderr,"Error in replacement_operation\n");
   fatal_error(1);
}
//----------------------------------------------------------------
if (is_equal_or_case_equal(text_letter,dic_letter,alph)) {
   // we are in the case of "london" instead of "London"
   return OP_REPLACEMENT_UPPERCASE;
}
//----------------------------------------------------------------
if (areSpellEquivalentLetters(text_letter,dic_letter)) {
   // we are in the case of "Lndon" instead of "London"
   return OP_REPLACEMENT_ACCENT;
}
//----------------------------------------------------------------
if (areCloseOnKeyboard(text_letter,dic_letter)) {
   // we are in the case of "zdvance" instead of "advance"
   return OP_REPLACEMENT_CLOSE_KEY;
}
return OP_REPLACEMENT;
}



int isReplacementOp(char c) {
   return (c==OP_REPLACEMENT)
           || (c==OP_REPLACEMENT_ACCENT)
           || (c==OP_REPLACEMENT_UPPERCASE)
           || (c==OP_REPLACEMENT_CLOSE_KEY);
}



//
// This function tests if the 2 last letters of 'text' and 'dic'
// have been switched. It returns INVALID_OP_SEQUENCE if there is
// a sequence operations that has produced a non switch like
// "c|ot|se" for "c|ro|se"; it returns VALID_OP_SEQUENCE otherwise
//
// Ex: "keyboadr" for "keyboard"
//
int check_letter_inversion(unichar* text,int text_pos,
                           unichar* dic,int dic_pos,
                           char* operations,int *n_op) {                           
if (text==NULL || dic==NULL) {
   // this case should not happen
   return VALID_OP_SEQUENCE;
}
if (text_pos==0 || dic_pos==0) {
   // if one of the words has less than 2 letters, we return
   return VALID_OP_SEQUENCE;
}
// first of all, we check if there is a real letter inversion
if (text[text_pos]==dic[dic_pos-1]
    && text[text_pos-1]==dic[dic_pos]
    && text[text_pos]!=text[text_pos-1]) {
    {
    unichar a=text[text_pos+1];
    text[text_pos+1]='\0';
    unichar b=dic[dic_pos+1];
    dic[dic_pos+1]='\0';
    text[text_pos+1]=a;
    dic[dic_pos+1]=b;
    }
    // then we check which operationshave produced this
    // inversion
    if ((*n_op) >=3
       && operations[(*n_op)-1]==OP_INSERTION
       && operations[(*n_op)-2]==OP_NORMAL
       && operations[(*n_op)-3]==OP_DELETION) {
       
       operations[*n_op]='\0';
       operations[(*n_op)]=OP_SWAP;
       (*n_op)++;
       operations[*n_op]='\0';
       return VALID_OP_SEQUENCE;
    }
    if ((*n_op) >=2
       && isReplacementOp(operations[(*n_op)-1])
       && isReplacementOp(operations[(*n_op)-2])) {

       operations[*n_op]='\0';
       operations[(*n_op)]=OP_SWAP;
       (*n_op)++;
       operations[*n_op]='\0';
       return VALID_OP_SEQUENCE;
   }
    if ((*n_op) >=3
       && operations[(*n_op)-1]==OP_DELETION
       && operations[(*n_op)-2]==OP_NORMAL
       && operations[(*n_op)-3]==OP_INSERTION) {

       operations[*n_op]='\0';
       operations[(*n_op)]=OP_SWAP;
       (*n_op)++;
       operations[*n_op]='\0';
       return VALID_OP_SEQUENCE;
    }
}
else {
   // If there is no letter swap, we check if there was an switch-like 
   // operation sequence anyway. In that case, we forbid this sequence
   // by returning INVALID_OP_SEQUENCE

    if ((*n_op) >=3
       && operations[(*n_op)-1]==OP_INSERTION
       && operations[(*n_op)-2]==OP_NORMAL
       && operations[(*n_op)-3]==OP_DELETION) {
       return INVALID_OP_SEQUENCE;
    }
    if ((*n_op) >=2
       && isReplacementOp(operations[(*n_op)-1])
       && isReplacementOp(operations[(*n_op)-2])) {
       return INVALID_OP_SEQUENCE;
   }
    if ((*n_op) >=3
       && operations[(*n_op)-1]==OP_DELETION
       && operations[(*n_op)-2]==OP_NORMAL
       && operations[(*n_op)-3]==OP_INSERTION) {
       return INVALID_OP_SEQUENCE;
    }

}
return VALID_OP_SEQUENCE;
}




//
// Checks if the current position in the input text is
// long enough to increment the error threshold. 
//
int update_number_of_errors_tolerated(int pos) {
return 0;
//return (pos%5==0)?1:0;
}




int exploreBinApproximate(int pos,unichar* word,int string_pos,unsigned char* BIN,
                int nErrors,
                unichar* correction,int c_pos,struct hypothesis_list** list,
                Alphabet* alph,char* operations,int n_op,int mode,
                int start_pos) {
int n_transitions;

// if there is a sequence that could have produced a letter switch and did not,
// we discard it.
// Ex: "c|ot|se" suggested for "c|ro|se"  => DISCARD
//     "c|or|se" suggested for "c|ro|se"  => ACCEPT
if (INVALID_OP_SEQUENCE==check_letter_inversion(word,string_pos-1,correction,c_pos-1,operations,&n_op)) {
   return 0;
}




int yuyu=0;
// we explore removing a letter: abc -> bc
if (nErrors>0 && is_removable_char(word[string_pos]) 
    // we do not delete after an insertion, because it would be
    // equivalent to a replacement
    && (n_op==0 || operations[n_op-1]!=OP_INSERTION)
    // we do not want to allow two consecutive deletions
    && (n_op==0 || operations[n_op-1]!=OP_DELETION)) {
   operations[n_op]=OP_DELETION;
   yuyu=yuyu+exploreBinApproximate(pos,word,string_pos+1,BIN,
                             nErrors-1+update_number_of_errors_tolerated(string_pos+1),
                             correction,c_pos,
                             list,alph,
                             operations,n_op+1,mode,start_pos);
}

n_transitions=((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1];
pos=pos+2;
//--------------------- BEGIN FINAL STATE IN DIC ----------------------
// First we check if we are at the end of text. It concerns both
// simple and compound words
int final=(!(n_transitions & 32768));
if (word[string_pos]=='\0') {
   // if we are at the end of the word, we check
   // if the state is final, i.e. if the word belongs
   // to the dictionary.
   if (final /*&& c_pos>SPELL_MINIMUM_SIZE_FOR_APPROXIMATE_MATCHING-2*/) {
      // if we have a word, we add it to the hypothesis list
      if (mode==SIMPLE_WORDS || (string_pos>mode && is_letter(word[string_pos-1],alph)
                                                 && !is_letter(word[string_pos],alph))) {
         // if it is a compound word, it must be longer than its first
         // simple word constituant, and it must end by a letter
         correction[c_pos]='\0';
         operations[n_op]='\0';
         int distance=0;
         int trust_value=0;
         if (!is_equal_ignore_case(correction,word,alph)) {
            // we compute a distance if and only if there a difference between
            // the word and its correction, because we can obtain a correction
            // identical to the word with several operations
            compute_distance_and_trust_value(operations,&distance,&trust_value);
         }
         if (list!=NULL) {
            //printf("string_pos=%d   len=%d\n_",string_pos,u_strlen(word));
            //u_prints(correction);printf("_");getchar();
            insert_in_hypothesis_list(list,
                new_spell_check_hypothesis(start_pos,
                                           start_pos+string_pos-1,
                                           (mode==SIMPLE_WORDS)?SPELL_ERROR_SIMPLE_WORD:((distance>0)?SPELL_ERROR_COMPOUND_WORD:SPELL_CORRECT_COMPOUND_WORD),
                                           distance,
                                           trust_value,
                                           word,
                                           correction,alph));
         }
      }
      yuyu=yuyu+1;
   }
   // we do not return here, because it would block the case
   // of an insertion at the end of a word
}
else if (final && mode!=SIMPLE_WORDS && string_pos>mode
         && is_letter(word[string_pos-1],alph)
         && !is_letter(word[string_pos],alph)) {
         // if we have a compound word
         unichar a=word[string_pos];
         word[string_pos]='\0';
         correction[c_pos]='\0';
         operations[n_op]='\0';
         int distance=0;
         int trust_value=0;
         if (!is_equal_ignore_case(correction,word,alph)) {
            // we compute a distance if and only if there a difference between
            // the word and its correction, because we can obtain a correction
            // identical to the word with several operations
            compute_distance_and_trust_value(operations,&distance,&trust_value);
         }
                  

         if (list!=NULL) {
            struct spell_check_hypothesis* h=new_spell_check_hypothesis(start_pos,
                                           start_pos+string_pos-1,
                                           (mode==SIMPLE_WORDS)?SPELL_ERROR_SIMPLE_WORD:((distance>0)?SPELL_ERROR_COMPOUND_WORD:SPELL_CORRECT_COMPOUND_WORD),
                                           distance,
                                           trust_value,
                                           word,
                                           correction,alph);
            insert_in_hypothesis_list(list,h);                               
         }
         word[string_pos]=a;
         //if (word[string_pos]=='\0') {return 1;}
}
//--------------------- END FINAL STATE IN DIC ----------------------

if ((n_transitions & 32768)) {
   // if we are in a normal node, we remove the control bit to
   // have the good number of transitions
   n_transitions=n_transitions-32768;
} else {
  // if we are in a final node, we must jump after the reference to the INF line number
  pos=pos+3;
}

for (int i=0;i<n_transitions;i++) {
  unichar c=(unichar)(((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1]);
  correction[c_pos]=c;
  pos=pos+2;
  int adr=((unsigned char)BIN[pos])*256*256+((unsigned char)BIN[pos+1])*256+(unsigned char)BIN[pos+2];
  pos=pos+3;
  // we explore normally: abc -> abc
  if (is_equal_or_case_equal(c,word[string_pos],alph)
      || are_equivalent_quotes(c,word[string_pos])) {
     operations[n_op]=OP_NORMAL;
     yuyu=yuyu+exploreBinApproximate(adr,word,string_pos+1,BIN,
                                     nErrors+update_number_of_errors_tolerated(string_pos+1),
                                     correction,c_pos+1,list,alph,
                                     operations,n_op+1,mode,start_pos);
  }
  // we explore replacing a letter: abc -> zbc
  else if (nErrors>0 && c!=word[string_pos] && can_replace_chars(word[string_pos],c,alph)
           // as the sequences deletion-replacement and replacement-deletion
           // are equivalent, we forbid one of them to avoid duplicates
           && (n_op==0 || operations[n_op-1]!=OP_DELETION)
           // as the sequences insertion-replacement and replacement-insertion
           // are equivalent, we forbid one of them to avoid duplicates
           && (n_op==0 || operations[n_op-1]!=OP_INSERTION)
           // we do not allow two consecutive replacement since
           // it is equivalent to insertion-normal-deletion
           && (n_op==0 || !isReplacementOp(operations[n_op-1]))
           // we do not allow a replacement after a sequence insertion-normal
           && ((n_op<2 || operations[n_op-1]!=OP_NORMAL || operations[n_op-2]!=OP_INSERTION) ||areSpellEquivalentLetters(word[string_pos],c)   )) {
          operations[n_op]=replacement_operation(word[string_pos],c,alph);
          yuyu=yuyu+exploreBinApproximate(adr,word,string_pos+1,BIN,
                                     nErrors-1+update_number_of_errors_tolerated(string_pos+1),
                                     correction,c_pos+1,list,alph,
                                     operations,n_op+1,mode,start_pos);
     }
  // we explore inserting a letter: abc -> yabc
  if (nErrors>0
      // we do not insert after a deletion, because it would be
      // equivalent to a replacement
      && (n_op==0 || operations[n_op-1]!=OP_DELETION)
      // we do not want to insert the same letter than the current one,
      // because this case is handled by the duplication case
      && (c!=word[string_pos])
      // we do not want to allow two consecutive insertions
      && (n_op==0 || operations[n_op-1]!=OP_INSERTION)
      // we do not allow an insertion after a sequence replace-normal
      && (n_op<2 || operations[n_op-1]!=OP_NORMAL || !isReplacementOp(operations[n_op-2]))) {
     // if that case, we do not call the 'update_number_of_errors_tolerated'
     // function, because we would increment nErrors several times
     // for the same length of explored text
     if (c_pos>0 && c==correction[c_pos-1]) {
        operations[n_op]=OP_DUPLICATION;
     } 
     else {operations[n_op]=OP_INSERTION;}
     yuyu=yuyu+exploreBinApproximate(adr,word,string_pos,BIN,nErrors-1,
                                     correction,c_pos+1,list,alph,
                                     operations,n_op+1,mode,start_pos);
  }
}
return yuyu;
}





int isCompoundWordModuloADistance(unichar* w,unsigned char* bin,
                                         int distance,
                                         struct hypothesis_list** list,
                                         Alphabet* alph,int mode,
                                         int start_pos) {
if (w==NULL || bin==NULL) {
   return 0;
}
unichar correction[1024];
char operations[1024];
return exploreBinApproximate(4,w,0,bin,distance,correction,0,list,
                             alph,operations,0,mode,start_pos);
}



int isSpellEquivalentWordModuloADistance(unichar* w,unsigned char* bin,
                                         int distance,
                                         struct hypothesis_list** list,
                                         Alphabet* alph,int start_pos) {
if (w==NULL || bin==NULL) {
   return 0;
}
unichar correction[1024];
char operations[1024];
return exploreBinApproximate(4,w,0,bin,distance,correction,0,list,alph,operations,0,SIMPLE_WORDS,start_pos);
}




int exploreBinInUpperCase(int pos,unichar* word,int string_pos,unsigned char* BIN,
                int nErrors,
                unichar* correction,int c_pos,struct hypothesis_list** list,
                Alphabet* alph,int start_pos) {
int n_transitions;
n_transitions=((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1];
pos=pos+2;
if (word[string_pos]=='\0') {
   // if we are at the end of the word, we check
   // if the state is final, i.e. if the word belongs
   // to the dictionary.
   if (!(n_transitions & 32768) /*&& c_pos>SPELL_MINIMUM_SIZE_FOR_APPROXIMATE_MATCHING-2*/) {
      // if we have a word, we add it to the hypothesis list
      correction[c_pos]='\0';
      insert_in_hypothesis_list(list,
                new_spell_check_hypothesis(start_pos,
                                           start_pos+string_pos-1,
                                           SPELL_ERROR_SIMPLE_WORD,
                                           ALL_IN_UPPER_CASE_DISTANCE,
                                           ALL_IN_UPPER_CASE_TRUST_VALUE,
                                           word,
                                           correction,alph));
      return 1;
   }
   return 0;
}
if ((n_transitions & 32768)) {
   // if we are in a normal node, we remove the control bit to
   // have the good number of transitions
   n_transitions=n_transitions-32768;
} else {
  // if we are in a final node, we must jump after the reference to the INF line number
  pos=pos+3;
}
for (int i=0;i<n_transitions;i++) {
  unichar c=(unichar)(((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1]);
  correction[c_pos]=c;
  pos=pos+2;
  int adr=((unsigned char)BIN[pos])*256*256+((unsigned char)BIN[pos+1])*256+(unsigned char)BIN[pos+2];
  pos=pos+3;
  // we explore normally: abc -> abc
  if (is_equal_or_case_equal(word[string_pos],c,alph)) {
     if (exploreBinInUpperCase(adr,word,string_pos+1,BIN,
                                     nErrors+update_number_of_errors_tolerated(string_pos+1),
                                     correction,c_pos+1,list,alph,start_pos)) {
        return 1;
     }
  }
}
return 0;
}






//
// Checks if 'w' is in the dictionary all in upper case.
// Ex: 'fbi' for 'FBI'
//
int isSpellEquivalentInUpperCase(unichar* w,unsigned char* bin,
                                         int distance,
                                         struct hypothesis_list** list,
                                         Alphabet* alph,int start_pos) {
if (w==NULL || bin==NULL) {
   return 0;
}
unichar correction[1024];
return exploreBinInUpperCase(4,w,0,bin,distance,correction,0,list,alph,start_pos);
}



//
// This function checks if w is a word in any dictionary of dic_list. If not,
// it checks if it a variant of a correct word of any dictionary of dic_list.
//
int getSimpleWordHypothesis(unichar* w,struct bin_list* dic_list,
                            Alphabet* alph,int distance,int start_pos,
                            struct hypothesis_list** global_list) {
struct hypothesis_list* list=NULL;
int there_are_suggestions=0;
for (int i=0;i<dic_list->n;i++) {
   if (isInDictionary(w,dic_list->tab[i]->bin,alph,SIMPLE_WORDS,NULL,start_pos)) {
      // if w is a correct simple word, we do not need to 
      // build any hypothesis
      return 0;
   }
}
// Then, we look if w is a correct word letter modifications
for (int i=0;i<dic_list->n;i++) {
   if (SPELL_LANGUAGE==SPELL_FRENCH) {
      if (isQuoteContraction(w,dic_list->tab[i],alph,&list,start_pos)) {
         there_are_suggestions=1;
      }
   }
   if (isSpellEquivalentInUpperCase(w,dic_list->tab[i]->bin,distance,
                                            &list,alph,start_pos)) {
      there_are_suggestions=1;
   }
   if (isSpellEquivalentWordModuloADistance(w,dic_list->tab[i]->bin,distance,
                                            &list,alph,start_pos)) {
      // if w is a word
      there_are_suggestions=1;
   }
}
//display_hypothesis_list(list); getchar();
clean_hypothesis_list(&list);
if (!there_are_suggestions || list==NULL) {
   // if there is no suggestion or too bad suggestions
   insert_in_hypothesis_list(&list,
       new_spell_check_hypothesis(start_pos,
                                  start_pos+u_strlen(w)-1,
                                  SPELL_WRONG_WORD_WITH_NO_HYPOTHESIS,
                                  0,
                                  SPELL_EMPTY_TRUST_VALUE,
                                  w,
                                  NULL,alph));
}

//display_hypothesis_list(list);
concat_hypothesis_list(global_list,list);
return 1;
}




int index_of_first_non_letter(unichar* w,Alphabet* alph) {
int i;
for (i=0;is_letter(w[i],alph);i++);
return i;
}


int index_of_last_non_letter(unichar* w,Alphabet* alph) {
int i;
int l=u_strlen(w);
for (i=l-1;i>0 && is_letter(w[i],alph);i--);
return i;
}


int beginning_of_next_token(unichar* w,Alphabet* alph) {
if (!is_letter(w[0],alph)) {
   return 1;
}
else {
   return index_of_first_non_letter(w,alph);
}
}


//
// This function checks if 'w' begins by compound words of any dictionary
// of dic_list. It also checks if it a variant of a correct compound 
// word of any dictionary of dic_list.
// Ex: "fire brigade ..." => "fire brigade"
//
int getCompoundWordHypothesis(unichar* w,struct bin_list* dic_list,
                             Alphabet* alph,int distance,int start_pos,
                             struct string_hash* stop_words,
                             struct hypothesis_list** global_list,
                             int is_a_stop_word) {

struct hypothesis_list* list=NULL;
int mode=index_of_first_non_letter(w,alph);
int res=0;
// we look for correct compound even if the first token is a stop word,
// in order to discard false simple word errors like 
// "dela" in "au-dela"
for (int i=0;i<dic_list->n;i++) {
   if (isInDictionary(w,dic_list->tab[i]->bin,alph,mode,&list,start_pos)) {
      // if w is a correct compound word, we build an
      // hypothesis in order to discard false unknown word like
      // "priori" in "a priori"
      res++;
   }
}
// even if the compound starts with a stop word, we look if
// it is an error due to a replacement of space and '-'
for (int i=0;i<dic_list->n;i++) {
   if (isInDictionaryModuloSeparators(w,dic_list->tab[i]->bin,alph,mode,&list,start_pos)) {
      // if w is a correct compound word, we build an
      // hypothesis in order to discard false unknown word like
      // "priori" in "a priori"
      res++;
   }
}
if (is_a_stop_word) {
   // if we have a stop word, we do not look for mispelled compound
   // words
   clean_compound_hypothesis_list(&list,stop_words,alph);
   concat_hypothesis_list(global_list,list);
   return res;
}
// Then, we look if w is a correct compound word after modifications
for (int i=0;i<dic_list->n;i++) {
   if (isCompoundWordModuloADistance(w,dic_list->tab[i]->bin,distance,
                                            &list,alph,mode,start_pos)) {
      //printf("word with errors in dic %d\n",i);
      res++;
   }
}
clean_compound_hypothesis_list(&list,stop_words,alph);
concat_hypothesis_list(global_list,list);
return res;
}






int exploreBinExactly(int pos,unichar* word,int string_pos,unsigned char* BIN,
                      Alphabet* ALPH,int mode,struct hypothesis_list** list,
                      int start_pos) {

int n_transitions;
int return_value=0;
n_transitions=((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1];
pos=pos+2;
//------------------- BEGIN FINAL STATE IN DIC -------------------------
// we check the end of text for both simple and compound words
int final=(!(n_transitions & 32768));
if (word[string_pos]=='\0') {
   // if we are at the end of the word, we just check
   // if the state is final, i.e. if the word belongs
   // to the dictionary
   if (!final) {return 0;}
   if (mode!=SIMPLE_WORDS) {
      // in the case of a compound word, we must verify that it is a valid
      // word frontier, that is to say that the word was ended by a letter.
      // We also must check that the end of the word is over the limit of
      // its first simple word constituant.
      if (string_pos>mode && is_letter(word[string_pos-1],ALPH)) {
         if (list!=NULL) insert_in_hypothesis_list(list,
         new_spell_check_hypothesis(start_pos,
                                  start_pos+string_pos-1,
                                  SPELL_CORRECT_COMPOUND_WORD,
                                  0,
                                  SPELL_EMPTY_TRUST_VALUE,
                                  word,
                                  NULL,ALPH));
      } else {return 0;}
   }
   else {return 1;}
}
else if (final && string_pos>mode && is_letter(word[string_pos-1],ALPH)
                                  && !is_letter(word[string_pos],ALPH)) {
   unichar a=word[string_pos];
   word[string_pos]='\0';
   return_value=1;
   if (list!=NULL) {
      insert_in_hypothesis_list(list,
         new_spell_check_hypothesis(start_pos,
                                  start_pos+string_pos-1,
                                  SPELL_CORRECT_COMPOUND_WORD,
                                  0,
                                  SPELL_EMPTY_TRUST_VALUE,
                                  word,
                                  NULL,ALPH));
   }
   word[string_pos]=a;
}
//------------------- END FINAL STATE IN DIC -------------------------
// then we check if we are 
if ((n_transitions & 32768)) {
   // if we are in a normal node, we remove the control bit to
   // have the good number of transitions
   n_transitions=n_transitions-32768;
} else {
  // if we are in a final node, we must jump after the reference to the INF line number
  pos=pos+3;
}
for (int i=0;i<n_transitions;i++) {
  unichar c=(unichar)(((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1]);
  pos=pos+2;
  int adr=((unsigned char)BIN[pos])*256*256+((unsigned char)BIN[pos+1])*256+(unsigned char)BIN[pos+2];
  pos=pos+3;
  if (is_equal_or_case_equal(c,word[string_pos],ALPH)
      || are_equivalent_quotes(c,word[string_pos])) {
     if (exploreBinExactly(adr,word,string_pos+1,BIN,ALPH,mode,list,start_pos)) return 1;
  }
}
return return_value;
}



//
// This function tests if w is a word in the dictionary bin
//
int isInDictionary(unichar* w,unsigned char* bin,Alphabet* alph,int mode,
                   struct hypothesis_list** list,int start_pos) {
if (w==NULL || bin==NULL) return -1;
return exploreBinExactly(4,w,0,bin,alph,mode,list,start_pos);
}





//@@@@@@@@@@@@@ BEGIN COMPOUND SEPARATOR ERRORS @@@@@@@@@@@@@@@


int is_a_valid_separator_replacement(unichar a,unichar b) {
if (a==b) {
  return 0;
}
return ((a==' ' && b=='-') ||
    (a=='-' && b==' '));
}


int exploreBinModuloSeparators(int pos,unichar* word,int string_pos,unsigned char* BIN,
                      Alphabet* ALPH,int minimum_end,struct hypothesis_list** list,
                      int start_pos,int nErrors,unichar* correction,int c_pos) {

int n_transitions;
int return_value=0;
n_transitions=((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1];
pos=pos+2;
//------------------- BEGIN FINAL STATE IN DIC -------------------------
// we check the end of text for both simple and compound words
int final=(!(n_transitions & 32768));
if (word[string_pos]=='\0') {
   // if we are at the end of the word, we just check
   // if the state is final, i.e. if the word belongs
   // to the dictionary
   if (!final) {return 0;}
      // in the case of an end of compound word, we must verify that it is a valid
      // word frontier, that is to say that the word was ended by a letter.
      // We also must check that the end of the word is over the limit of
      // its first simple word constituant. We also check that there is actually
      // an error.
      if (nErrors>0 && string_pos>minimum_end && is_letter(word[string_pos-1],ALPH)) {
         correction[c_pos]='\0';
         if (list!=NULL) insert_in_hypothesis_list(list,
         new_spell_check_hypothesis(start_pos,
                                  start_pos+string_pos-1,
                                  SPELL_ERROR_COMPOUND_WORD,
                                  nErrors,
                                  SPELL_BONUS_COMPOUND_SEPARATOR_ERROR,
                                  word,
                                  correction,ALPH));
      } else {return 0;}
}
else if (nErrors>0 && final && string_pos>minimum_end && is_letter(word[string_pos-1],ALPH)
                                  && !is_letter(word[string_pos],ALPH)) {
   unichar a=word[string_pos];
   word[string_pos]='\0';
   return_value=1;
   if (list!=NULL) {
      correction[c_pos]='\0';
      insert_in_hypothesis_list(list,
         new_spell_check_hypothesis(start_pos,
                                  start_pos+string_pos-1,
                                  SPELL_ERROR_COMPOUND_WORD,
                                  nErrors,
                                  SPELL_BONUS_COMPOUND_SEPARATOR_ERROR,
                                  word,
                                  correction,ALPH));
   }
   word[string_pos]=a;
}
//------------------- END FINAL STATE IN DIC -------------------------
// then we check if we are 
if ((n_transitions & 32768)) {
   // if we are in a normal node, we remove the control bit to
   // have the good number of transitions
   n_transitions=n_transitions-32768;
} else {
  // if we are in a final node, we must jump after the reference to the INF line number
  pos=pos+3;
}
for (int i=0;i<n_transitions;i++) {
  unichar c=(unichar)(((unsigned char)BIN[pos])*256+(unsigned char)BIN[pos+1]);
  pos=pos+2;
  int adr=((unsigned char)BIN[pos])*256*256+((unsigned char)BIN[pos+1])*256+(unsigned char)BIN[pos+2];
  pos=pos+3;
  if (is_equal_or_case_equal(c,word[string_pos],ALPH)
      || are_equivalent_quotes(c,word[string_pos])) {
     correction[c_pos]=c;
     if (exploreBinModuloSeparators(adr,word,string_pos+1,BIN,ALPH,minimum_end,list,start_pos,nErrors,correction,c_pos+1)) return 1;
  }
  else if (is_a_valid_separator_replacement(c,word[string_pos])) {
          // if we can replace a separator by another, then we do it
          correction[c_pos]=c;
          if (exploreBinModuloSeparators(adr,word,string_pos+1,BIN,ALPH,minimum_end,list,start_pos,nErrors+1,correction,c_pos+1)) return 1;
  }
}
return return_value;
}



//
// This function tests if w is a compound word in the dictionary bin,
// with a tolerance for separators switches.
// Ex: if the dictionary contains "au del", the word "au-del" will be
//     accepcted
//
int isInDictionaryModuloSeparators(unichar* w,unsigned char* bin,Alphabet* alph,int mode,
                   struct hypothesis_list** list,int start_pos) {
if (w==NULL || bin==NULL) return -1;
unichar correction[1024];
return exploreBinModuloSeparators(4,w,0,bin,alph,mode,list,start_pos,0,correction,0);
}

//@@@@@@@@@@@@@ END COMPOUND SEPARATOR ERRORS @@@@@@@@@@@@@@@

